import warnings

import pytest

import anndata

# TODO: Should be done in pyproject.toml, see anndata/conftest.py
warnings.filterwarnings("ignore", category=anndata.OldFormatWarning)


@pytest.fixture
def backing_h5ad(tmp_path):
    return tmp_path / "test.h5ad"
